/*
 * oscp-slave-timi-libtimidity-0.1.0.c
 * Copyright (C) 2015 KotCzarny <tjosko@yahoo.com>
 *
 *  This program is free software; you can redistribute it and/or modify it
 *  under the terms of the GNU General Public License as published by the
 *  Free Software Foundation; either version 2 of the License, or (at your
 *  option) any later version.
 *
 *  This program is distributed in the hope that it will be useful, but
 *  WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License along
 *  with this program; if not, write to the Free Software Foundation, Inc.,
 *  59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 */

#include "oscp-slave.h"

#include <timidity.h>

MidIStream *timic;
MidSongOptions timio;
MidSong *timis;
MidDLSPatches *timip;

#define timi_dbuf_size 1024*2
char timi_dbuf[timi_dbuf_size];

int timistate=0; float timicur=0, timidur=0; int timii=0;
int timisong=0; int timidurmax=5;

/////################
void timi_show_supp() {
 printf("\n  [slave-timi]: based on libtimidity-0.1.0");
 printf(", formats: mid\n");
}

/////###########
//void timi_flush() { lf("timi_flush"); adev_flush(); }
/////###########
void timi_init() { lf("timi_init"); timistate=0; dlog(9, "timi_init"); if(timistate>0) return;

// if(timic) dlog(9, "timi_init: timic not null?");

 if(adev_init()<=0) { return; }

// if(mid_init(NULL)<0) { dlog(9, "timi_init: mid_init failed"); return; }
 if(mid_init_no_config()<0) { dlog(9, "timi_init: mid_init failed"); return; }

 timio.rate = 44100;
 timio.format = MID_AUDIO_S16LSB;
 timio.channels = 2;
 timio.buffer_size = timi_dbuf_size/4;

// char *patchsetfn="8mbgm_enhanced18.dls";
// char *patchsetfn="gm.dls";
// char *patchsetfn="Fury.dls";
 char *patchsetfn="sf/0.dls";
 int timifs=get_filesize(patchsetfn);
 if(timifs<=0) { dlog(9,"timi_init: invalid patchset size"); return; }

 char *patchset=malloc(timifs); if(!patchset) { dlog(9,"timi_init: malloc failed (patchset)\n"); return; }
 FILE *f=fopen(patchsetfn, "r"); if(!f) { dlog(9,"timi_init: fopen failed (patchset)\n"); return; }
 fread(patchset, timifs, 1, f); fclose(f);

// stream = mid_istream_open_fp (stdin, 0);
// stream = mid_istream_open_file (argv[arg]);
 timic=mid_istream_open_mem(patchset, timifs, 0);
 timip=mid_dlspatches_load(timic); mid_istream_close(timic); free(patchset);

 timistate=1;
}
/////###########
void timi_stop() { lf("timi_stop");
 dlog(9, "timi_stop");
 if(timistate<=1) return;

 if(timistate>=3) { 
  mid_song_free(timis);
 }

 timistate=1;
}
/////##########
void timi_end() { lf("timi_end");
 dlog(9, "timi_end"); if(timistate>1) timi_stop(); adev_end();

 mid_dlspatches_free(timip);
 mid_exit();

 timistate=0; 
}
/////###################
void timi_die(char *msg) { lf2("timi_die",msg); dlog(9, "timi_die"); timi_stop(); cs(msg); }

char timifn[PATH_MAX]; int timifs=0;

/////##########################
void timi_update_info(char *fn) { lf("timi_update_info");
 static char lmsg[4096];

 snprintf(lmsg, 4095, "Title: %s (%s)\n",
  mid_song_get_meta(timis, MID_SONG_TEXT), get_basename(fn)); cs(lmsg);

 timidur=mid_song_get_total_time(timis); if(timidur<=0) timidur=timidurmax;
 sprintf(lmsg, "ANS_length=%.1f\n", timidur/1000); cs(lmsg);

 int fs=get_filesize(fn);
 snprintf(lmsg,1023,"Comment: %d,%s\n", fs, "MID"); cs(lmsg);
}

void timi_play1();
/////################################
void timi_load(char *input_filename0) { lfd("timi_load:state",timistate);
 static char lmsg[4096]; lmsg[4095]=0;
 snprintf(lmsg, 4095, "Playing %s\n", input_filename0); cs(lmsg);

 if(input_filename0[0]==0) return;
 if(input_filename0[0]==*"\"" && input_filename0[1]==*"\"" && input_filename0[2]==0)
 { cs("EOF code: stop requested\n"); if(timistate>2) timi_stop(); return; }

 if(timistate==0) timi_init();
 else if(timistate>2) timi_stop();

 timistate=2; timicur=0; timidur=0;
 char *input_filename=input_filename0;


 timifs=get_filesize(input_filename);
 if(timifs<=0) { timi_die("EOF code: invalid module size\n"); return; }

 char *module=malloc(timifs); if(!timifs) { timi_die("EOF code: malloc failed\n"); return; }
 FILE *f=fopen(input_filename, "r"); if(!f) { timi_die("EOF code: fopen failed\n"); return; }
 fread(module, timifs, 1, f); fclose(f);

// stream = mid_istream_open_fp (stdin, 0);
// stream = mid_istream_open_file (argv[arg]);
 timic=mid_istream_open_mem(module, timifs, 0);
// timis=mid_song_load(timic, &timio);
 timis=mid_song_load_dls(timic, timip, &timio);
//MidSong *mid_song_load_dls (MidIStream * stream, MidDLSPatches * patches, MidSongOptions * options);

 mid_istream_close(timic); free(module);
 if(timis==NULL) { timi_die("EOF code: mid_song_load failed\n"); timi_stop(); return; }

 mid_song_set_volume(timis, 100);
 mid_song_start(timis);

// mid_song_free (song); mid_exit ();
//      char *title = mid_song_get_meta (song, MID_SONG_TEXT);
//  if (title == NULL) { if (arg < argc) title = argv[arg]; else title = "stdin"; }
//  buffer_size = bits * channels / 8 * rate;
//  buffer = malloc (buffer_size);
//  long time = mid_song_get_time (song);  printf ("\rTime: %02ld:%02ld/%02ld:%02ld",
//   time / 60000, (time / 1000) % 60, total_time / 60000, (total_time / 1000) % 60);
//  bytes_read = mid_song_read_wave (song, buffer, buffer_size);
//  ao_play (device, buffer, bytes_read);
//  while (bytes_read);  free (buffer);
//  ao_close (device);  ao_shutdown ();
//  mid_song_free (song);  mid_exit ();


//  format.byte_format = AO_FMT_LITTLE;
 sfmt=AV_SAMPLE_FMT_S16;
      if(sfmt==AV_SAMPLE_FMT_U8 || sfmt==AV_SAMPLE_FMT_U8P){ sformat.bits=8; }
 else if(sfmt==AV_SAMPLE_FMT_S16 || sfmt==AV_SAMPLE_FMT_S16P){ sformat.bits=16; }
 else if(sfmt==AV_SAMPLE_FMT_S32 || sfmt==AV_SAMPLE_FMT_S32P){ sformat.bits=32; }
 else if(sfmt==AV_SAMPLE_FMT_FLT || sfmt==AV_SAMPLE_FMT_FLTP){ sformat.bits=32; }
 else if(sfmt==AV_SAMPLE_FMT_DBL || sfmt==AV_SAMPLE_FMT_DBLP){ sformat.bits=64; }
 else { timi_die("EOF code: Unsupported format\n"); return; }

 timicur=-1;
// timidur=timiInfo_GetDuration(timiI, timiInfo_GetDefaultSong(timiI))/1000.0;

 sformat.channels=2; sformat.rate=44100;
#ifdef LAO
 sformat.byte_format=AO_FMT_NATIVE; sformat.matrix=0;
#endif

 adev_setp(sformat.bits, &sformat.channels, &sformat.rate);
sprintf(lmsg, "AUDIO: %d Hz, %d ch, xxx, %d kbit\n", sformat.rate, sformat.channels, 0); cs(lmsg);

 timi_update_info(input_filename);

 timistate=3; timii=1;
 timi_play1();
}
void oscp_timi_check_input();
void oscp_timi_parse(char *a);
/////################# check messages
void timi_check_input() { lf("timi_check_input"); static char lmsg[4096]; static int rc;
 static fd_set timifsa;  static struct timeval timitv,*tvp;
 if(timistate==4) { tvp=NULL; }
 else { timitv.tv_sec=0; timitv.tv_usec=1; tvp=&timitv; }

 FD_ZERO(&timifsa); if(parent_said>=0) FD_SET(parent_said, &timifsa);

 rc=select(FD_SETSIZE, &timifsa, NULL, NULL, tvp); if(rc==0) return;
 else if(rc<0) { sprintf(lmsg, "Info: timici rc<0, %d, %s\n", errno, strerror(errno)); cs(lmsg); }
 else if(rc>0) // check if input is available
 { fdreadline(parent_said, lmsg, 1020); oscp_timi_parse(lmsg); }
}
/////###################
void timi_pause_toggle() { lf("timi_pause_toggle");
      if(timistate==3) { timistate=4; cs("===== PAUSE =====\n");
  do { oscp_timi_check_input(); } while(timistate==4);
 }
 else if(timistate==4) { timistate=3; cs("===== RESUME =====\n"); }
 else cs("Info: pause called but timistate != (3 || 4)"); }

/////##############
void timi_checksd() {}
/////##############
void timi_checkmd() {}

/////############
void timi_play1() { lf("timi_play1"); dlog(9, "timi_play1");
 if(timistate!=3) { cs("EOF code: can't play, not inited\n"); return; }
 float ot1=-7.0, ts=((float)timi_dbuf_size)/timio.rate/timio.channels/2.0;

 static char lmsg[128]; void *buf[1]; buf[0]=timi_dbuf;
 // short int *buf2=(short int *)timi_dbuf; char *gerr=NULL;
 int trimsilence=1;
 int br;

//FILE *f=fopen("oscp-timi.debug","a");
 while(timistate==3 && (br=mid_song_read_wave(timis,timi_dbuf,timi_dbuf_size))>0
 && timicur<=timidur )
 {
  timicur+=ts; if((int)timicur!=(int)ot1)
  { ot1=timicur; sprintf(lmsg, "A: %.1f 0 0 0 0 0 0 0\n", timicur); cs(lmsg); }
//sprintf(lmsg, "timi_play1: br: %d, timi_dbuf_size: %d", br, timi_dbuf_size);dlog(9, lmsg);
//fwrite(timi_dbuf, br, 1, f);

if(trimsilence==1) { int i; for(i=0;i<br;i++) if(timi_dbuf[i]!=0) { trimsilence=0; break; }}
if(trimsilence==0)
  adev_out(buf, br/2/sformat.channels, sformat.channels, sfmt);

  timii++; timi_check_input();
 }

 adev_close();
 timi_die("EOF code: finished decoding\n");
 lf("timi1_play:finished");
}

/////########################
void oscp_timi_load(char *fn) { lf("oscp_timi_load");
 timi_load(fn); //timi1_play();
 lf("oscp_timi_load:finished");
}
/////#####################
void oscp_timi_seek(int a) { lf("oscp_timi_seek"); int t=timicur+a; if(t<0) t=0;
// timi_seek(timic, t*1000.0); timicur=timi_tell(timic)/1000;
//   extern void mid_song_seek (MidSong * song, uint32 ms);
}
/////########################
void oscp_timi_parse(char *a) { lf("oscp_timi_parse"); char lmsg[4096];
// sprintf(lmsg,"ofparse start"); dlog(9,lmsg);
 int l=strlen(a); if(l<=0) { sprintf(lmsg,"ofparse: l==%d ?", l); dlog(9,lmsg); return; }
 if(strcmp(a,"q")==0 || strcmp(a,"quit")==0) { timi_end(); exit(0);
 } else if(strncmp(a,"loadfile ",9)==0) { oscp_timi_load(a+9);
 } else if(strncmp(a,"load ",5)==0) { oscp_timi_load(a+5);
 } else if(strncmp(a,"pause",5)==0) { timi_pause_toggle();
 } else if(strncmp(a,"get_property length",19)==0) {
  sprintf(lmsg, "ANS_length=%.1f\n", timidur); cs(lmsg);
 } else if(strncmp(a,"get_property time_pos",19)==0 || strncmp(a,"get_time_pos",12)==0) {
//  sprintf(lmsg, "A: %.1f 0 0 0 0 0 0 0 0 0\n", timicur); cs(lmsg);
 } else if(strncmp(a,"seek ",5)==0) { int o1=0;
  if(a[5]==*"-") { sscanf(a+6, "%d", &o1); o1*=-1; }
  else if(a[5]==*"+") { sscanf(a+6, "%d", &o1); }
  else { sscanf(a+5, "%d", &o1); }
  oscp_timi_seek(o1);
 } else if(strncmp(a,"vis ",4)==0) { sscanf(a+4, "%d", &vis);
 } else if(strncmp(a,"lfdump",5)==0) { lfdump();
 } else { sprintf(lmsg, "ofparse unhandled: %s", a); dlog(9, lmsg); }
 lf("oscp_timi_parse:finished");
}
/////####################### check stdin
void oscp_timi_check_input() { lf("oscp_timi_check_input");
 char lmsg[4096]; int i, rc; fd_set fsa;
 struct timeval *tvp=NULL; //, tv; tv.tv_sec=0; tv.tv_usec=10000*99; // i=(int)time(NULL);
 usleep(10000); FD_ZERO(&fsa); if(parent_said>=0) FD_SET(parent_said, &fsa);
 rc=select(FD_SETSIZE, &fsa, NULL, NULL, tvp); // block till input

 if(rc<0) { if(0) { exit(EXIT_FAILURE); } else usleep(10000*99); }
 else if(rc==0) { }
 else for(i=0; i<FD_SETSIZE; ++i) if(FD_ISSET(i, &fsa)) {
  if(i==parent_said) { fdreadline(i, lmsg, 1020); oscp_timi_parse(lmsg); }
 }
 lf("oscp_timi_check_input:finished");
}
////################
int oscp_timi_main() { lf("oscp_timi_main");
 cs("oscp-slave libtimidity-0.1.0\n");
 lf("oscp_timi_main:timi_init");
 timi_init();
 lf("oscp_timi_main:while");
 while(1) { oscp_timi_check_input(); }
 lf("oscp_timi_main:timi_end");
 timi_end();
 lf("oscp_timi_main:finished");
 cs("EOF code: exit\n");
// lfdump();
 return 0;
}
